"""
WSGI interface.

The wsgi/http server is inspired in `Tornado <http://www.tornadoweb.org/>`_
and `Waitress <http://docs.pylonsproject.org/projects/waitress>`_ projects
and using the best of both.
"""
