# -*- encoding: utf-8 -*-
"""
A python library to do development based in Finite State Machines.
"""
__version__ = '0.6.8'
__title__ = 'ginsfsm'
__author__ = 'Ginés Martínez'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2012-2013 Ginés Martínez'

