""" Inter event example.

.. image:: inter_event_schema.png


See the examples in directories:

* ginsfsm.examples.inter_event.ap1_banco
* ginsfsm.examples.inter_event.ap2_financiera
* ginsfsm.examples.inter_event.ap3_bolsa
* ginsfsm.examples.inter_event.ap4_publicidad

"""
